import json
from pathlib import Path
import logging
from itertools import chain
from importlib import resources
from . import schemas

from jsonschema import Draft202012Validator
from referencing import Registry, Resource
from referencing.jsonschema import DRAFT202012


class JsonValidator:
    def __init__(self, core_schema, other_schemas):
        self._core_schema = json.load(core_schema.open())
        self._schema_dict = {self._core_schema["$id"]: self._core_schema}
        self._supporting_paths = []
        for thing in other_schemas:
            path = Path(thing)
            if path.is_dir():
                logging.debug(f"Searching {path} for schema files")
                self._supporting_paths.extend(path.rglob("*.schema.json"))
            else:
                logging.debug(f"Appending {path} as schema file")
                self._supporting_paths.append(path)
        logging.info(f"Supporting schema paths: {self._supporting_paths}")
        self._supporting_schemas = [json.load(p.open()) for p in self._supporting_paths]
        self._schema_dict = self._schema_dict | {
            s["$id"]: s for s in self._supporting_schemas
        }
        logging.info(f"Loaded schema IDs: {[k for k in self._schema_dict.keys()]}")
        # self._resolver = RefResolver(None,
        #                 referrer=None,
        #                 store=self._schema_dict)
        # logging.info("Created RefResolver")
        self._registry = Registry().with_resources(
            [
                (k, Resource(contents=v, specification=DRAFT202012))
                for k, v in self._schema_dict.items()
            ]
        )
        self._validator = Draft202012Validator(
            self._core_schema, registry=self._registry
        )
        logging.info("Created validator")

    def validate(self, instance_doc: str):
        errors = list(self._validator.iter_errors(instance_doc))
        return errors


class TS103976Validator(JsonValidator):
    def __init__(self):
        repo_path = resources.files(schemas).joinpath("dummy").parent  # Hack
        schema_dirs = [repo_path]
        core_schema = repo_path / "ts_103976.schema.json"
        JsonValidator.__init__(self, core_schema, schema_dirs)

class TS103120Validator(JsonValidator):
    def __init__(self):
        repo_path = resources.files(schemas).joinpath("dummy").parent  # Hack
        schema_dirs = [repo_path]
        core_schema = repo_path / "ts_103120_Core.schema.json"
        JsonValidator.__init__(self, core_schema, schema_dirs)
        request_fragment_schema = {
            "$ref": "ts_103120_Core_2019_10#/$defs/RequestPayload"
        }
        self._request_fragment_validator = Draft202012Validator(
            request_fragment_schema, registry=self._registry
        )
        response_fragment_schema = {
            "$ref": "ts_103120_Core_2019_10#/$defs/ResponsePayload"
        }
        self._response_fragment_validator = Draft202012Validator(
            response_fragment_schema, registry=self._registry
        )

    def expand_request_response_exception(self, ex):
        if list(ex.schema_path) == ["properties", "Payload", "oneOf"]:
            logging.info(
                "Error detected validating payload oneOf - attempting explicit validation..."
            )
            if "RequestPayload" in self.instance_doc["Payload"].keys():
                ret_list = list(
                    chain(
                        *[
                            self.expand_action_exception(x)
                            for x in self._request_fragment_validator.iter_errors(
                                self.instance_doc["Payload"]["RequestPayload"]
                            )
                        ]
                    )
                )
                for r in ret_list:
                    r.path = ex.path + r.path
                return ret_list
            elif "ResponsePayload" in self.instance_doc["Payload"].keys():
                ret_list = list(
                    chain(
                        *[
                            self.expand_action_exception(x)
                            for x in self._request_fragment_validator.iter_errors(
                                self.instance_doc["Payload"]["ResponsePayload"]
                            )
                        ]
                    )
                )
                for r in ret_list:
                    r.path = ex.path + r.path
                return ret_list
            else:
                logging.error(
                    "No RequestPayload or ResponsePayload found - is the Payload malformed?"
                )
                return [ex]
        else:
            return [ex]

    def expand_action_exception(self, ex):
        logging.error("Error detected in ActionRequests/ActionResponses")
        error_path = list(ex.schema_path)
        if error_path != [
            "properties",
            "ActionRequests",
            "properties",
            "ActionRequest",
            "items",
            "allOf",
            1,
            "oneOf",
        ] and error_path != [
            "properties",
            "ActionResponses",
            "properties",
            "ActionResponse",
            "items",
            "allOf",
            1,
            "oneOf",
        ]:
            logging.error("Error not in inner Request/Response allOf/oneOf constraint")
            return [ex]
        j = ex.instance
        j.pop(
            "ActionIdentifier"
        )  # Remove ActionIdentifier - one remaining key will be the verb
        verb = list(j.keys())[0]
        message = "Request" if error_path[1] == "ActionRequests" else "Response"
        v = Draft202012Validator(
            {"$ref": f"ts_103120_Core_2019_10#/$defs/{verb}{message}"},
            registry=self._registry,
        )
        ret_list = list(
            chain(*[self.expand_object_exception(x) for x in v.iter_errors(j[verb])])
        )
        for r in ret_list:
            r.path = ex.path + r.path
        return ret_list

    def expand_object_exception(self, ex):
        logging.error("Error detected in verb")
        # The final level of validation is for the actual HI1Object validation
        if list(ex.schema_path) != ["properties", "HI1Object", "oneOf"]:
            logging.error("Error not inside HI1Object")
            return [ex]
        object_type = ex.instance["@xsi:type"].split("}")[-1]
        object_ref = {
            "AuthorisationObject": "ts_103120_Authorisation_2020_09#/$defs/AuthorisationObject",
            "LITaskObject": "ts_103120_Task_2020_09#/$defs/LITaskObject",
            "LDTaskObject": "ts_103120_Task_2020_09#/$defs/LDTaskObject",
            "LPTaskObject": "ts_103120_Task_2020_09#/$defs/LPTaskObject",
            "DocumentObject": "ts_103120_Document_2020_09#/$defs/DocumentObject",
            "NotificationObject": "ts_103120_Notification_2016_02#/$defs/NotificationObject",
            "DeliveryObject": "ts_103120_Delivery_2019_10#/$defs/DeliveryObject",
            "TrafficPolicyObject": "ts_103120_TrafficPolicy_2022_07#/$defs/TrafficPolicyObject",
            "TrafficRuleObject": "ts_103120_TrafficPolicy_2022_07#/$defs/TrafficRuleObject",
        }[object_type]
        v = Draft202012Validator({"$ref": object_ref}, registry=self._registry)
        print(ex.instance)
        print(object_ref)
        return list(v.iter_errors(ex.instance))

    def validate(self, instance_doc: str):
        errors = JsonValidator.validate(self, instance_doc)
        self.instance_doc = instance_doc
        out_errors = list(
            chain(*[self.expand_request_response_exception(ex) for ex in errors])
        )
        return out_errors
        return errors



from http.server import BaseHTTPRequestHandler, HTTPServer

class S(BaseHTTPRequestHandler):
    def _set_response(self):
        self.send_response(200)
        self.send_header('Content-type', 'application/json')
        self.end_headers()

    def do_GET(self):
        self._set_response()
        self.wfile.write("GET request for {}".format(self.path).encode('utf-8'))

    def do_POST(self):
        content_length = int(self.headers['Content-Length']) # <--- Gets the size of data
        post_data = self.rfile.read(content_length) # <--- Gets the data itself
        print(post_data.decode('utf-8'))
        self._set_response()
        # TODO be a proper HI1 endpoint
        self.wfile.write("POST request for {}".format(self.path).encode('utf-8'))

def run(server_class=HTTPServer, handler_class=S, port=8443):
    server_address = ('', port)
    httpd = server_class(server_address, handler_class)
    try:
        httpd.serve_forever()
    except KeyboardInterrupt:
        pass
    httpd.server_close()

if __name__ == '__main__':
    from sys import argv

    if len(argv) == 2:
        run(port=int(argv[1]))
    else:
        run()


from uuid import uuid4

from .helpers import etsi_datetime


class LEA:
    def __init__(self, country_code: str, unique_id: str, delivery_url: str = None):
        self.sender_country_code = country_code
        self.sender_unique_id = unique_id
        self.delivery_url = delivery_url

    def generate_103120_query(
        self, objects_to_create: list, receiver_country_code: str, receiver_unique_id
    ) -> dict:
        return {
            "Header": {
                "@xmlns": "http://uri.etsi.org/03120/common/2019/10/Core",
                "@xmlns:xsi": "http://www.w3.org/2001/XMLSchema-instance",
                "@xmlns:common": "http://uri.etsi.org/03120/common/2016/02/Common",
                "@xmlns:task": "http://uri.etsi.org/03120/common/2020/09/Task",
                "SenderIdentifier": {
                    "CountryCode": self.sender_country_code,
                    "UniqueIdentifier": self.sender_unique_id,
                },
                "ReceiverIdentifier": {
                    "CountryCode": receiver_country_code,
                    "UniqueIdentifier": receiver_unique_id,
                },
                "TransactionIdentifier": str(uuid4()),
                "Timestamp": etsi_datetime(),
                "Version": {
                    "ETSIVersion": "V1.16.1",
                    "NationalProfileOwner": "XX",
                    "NationalProfileVersion": "v1.0",
                },
            },
            "Payload": {
                "RequestPayload": {
                    "ActionRequests": {
                        "ActionRequest": [
                            {"ActionIdentifier": i, "CREATE": {"HI1Object": o}}
                            for i, o in enumerate(objects_to_create)
                        ]
                    }
                }
            },
        }

    def generate_empty_LDTaskRequest(
        self, task_ref: str, request_type: str, delivery_url: str = None
    ) -> dict:
        ret_dict = {
            "@xsi:type": "{http://uri.etsi.org/03120/common/2020/09/Task}LDTaskObject",
            "ObjectIdentifier": str(uuid4()),
            "CountryCode": self.sender_country_code,
            "OwnerIdentifier": self.sender_unique_id,
            "task:Reference": task_ref,
            "task:RequestDetails": {
                "task:Type": {
                    "common:Owner": "ETSI",
                    "common:Name": "TS103976RequestType",
                    "common:Value": request_type,
                },
                "task:RequestValues": {"task:RequestValue": []},
            }}
        if delivery_url is not None:
            ret_dict["task:DeliveryDetails"] = {
                  "task:DeliveryDestination": [
                    {
                      "task:DeliveryAddress": {
                        "task:URL": delivery_url
                      }
                    }
                  ]
            }
        return ret_dict

    def generate_vin_to_comms_id(
        self,
        receiver_country_code="XX",
        receiver_unique_id="Receiver",
        task_ref="XX-X-XXX",
        vin="1G9Y817H34LSP7293",
    ):
        ld_task = self.generate_empty_LDTaskRequest(
            task_ref, "VINtoCommsID", self.delivery_url
        )
        ld_task["task:RequestDetails"]["task:RequestValues"][
            "task:RequestValue"
        ].append(
            {
                "task:FormatType": {
                    "task:FormatOwner": "ETSI",
                    "task:FormatName": "VIN",
                },
                "task:Value": vin,
            }
        )
        hi1_request = self.generate_103120_query(
            [ld_task], receiver_country_code, receiver_unique_id
        )
        return hi1_request
    
    def generate_comms_id_to_vin(
        self,
        receiver_country_code="XX",
        receiver_unique_id="Receiver",
        task_ref="XX-X-XXX",
        imsi="999990123456789",
    ):
        ld_task = self.generate_empty_LDTaskRequest(
            task_ref, "CommsIDtoVIN", self.delivery_url
        )
        ld_task["task:RequestDetails"]["task:RequestValues"][
            "task:RequestValue"
        ].append(
            {
                "task:FormatType": {
                    "task:FormatOwner": "ETSI",
                    "task:FormatName": "IMSI",
                },
                "task:Value": imsi,
            }
        )
        hi1_request = self.generate_103120_query(
            [ld_task], receiver_country_code, receiver_unique_id
        )
        return hi1_request    

## implement comms_id_to_vin

##


# Implements a simple RPS
import logging

from uuid import uuid4
from datetime import datetime
from copy import deepcopy
from re import compile
import random
from hashlib import sha1
from dataclasses import dataclass
from typing import List
from json import dumps
from enum import Enum
from sys import stderr

from requests import post

from .validator import TS103120Validator
from .helpers import etsi_datetime



class DeliveryBehaviourEnum (Enum):
    SKIPDELIVERY = 0,
    FAIL = 1,
    DELIVERTODEFAULT = 2,
    WRITETOSTDOUT = 3,
    WRITETOSTDERR = 4
   
    # magic methods for argparse compatibility
    def __str__(self):
        return self.name.lower()

    def __repr__(self):
        return str(self)

    @staticmethod
    def from_string(s):
        try:
            return DeliveryBehaviourEnum[s.upper()]
        except KeyError:
            raise ValueError()        
        
@dataclass
class FulfilmentOutcome:
    errorCode: int = 0
    errorDescription: str = ""
    updatedObject: dict = None
    records: List = None

class RPS:
    SUPPORTED_COMMS_IDS = ["IMEI", "IMSI", "ICCID", "PEIIMEI", "SUPIIMSI", "SUPINAI", "MSISDN", "GPSIMSISDN", "GPSINAI", "MACAddress", "EUI64"]
    ALLOWED_VINS = ["1CPH423GA4G102745", "1G9Y817H34LSP7298", "WGV110000CMSP7891", "VFR1331AA40687997"]

    def __init__(
        self,
        ReceiverCC : str = "XX",
        ReceiverID : str = "RPSReceiver",
        AllowAnyReceiverID : bool = False,
        BehaviourIfNoDeliveryDest : DeliveryBehaviourEnum = DeliveryBehaviourEnum.SKIPDELIVERY,
        DefaultDeliveryDest : str = None
    ):
        self.receiverCC = ReceiverCC
        self.receiverID = ReceiverID
        self.allowAnyReceiverID = AllowAnyReceiverID
        self.behaviourIfNoDeliveryDest = BehaviourIfNoDeliveryDest
        self.defaultDeliveryDest = DefaultDeliveryDest
        self.validator = TS103120Validator()
        self.VIN_regex = compile(r"^[A-HJ-NPR-Z0-9]{17}$")
        self.deliveryTimeout = 5
        self.alwaysLogResults = True

    def _generate_top_level_error(self, error_message: str) -> dict:
        return {
            "@xmlns": "http://uri.etsi.org/03120/common/2019/10/Core",
            "@xmlns:xsi": "http://www.w3.org/2001/XMLSchema-instance",
            "@xmlns:common": "http://uri.etsi.org/03120/common/2016/02/Common",
            "Header": {
                "SenderIdentifier": {
                    "CountryCode": "ZZ",
                    "UniqueIdentifier": "UNKNOWN",
                },
                "ReceiverIdentifier": {
                    "CountryCode": self.receiverCC,
                    "UniqueIdentifier": self.receiverID,
                },
                "TransactionIdentifier": str(uuid4()),
                "Timestamp": etsi_datetime(),
                "Version": {
                    "ETSIVersion": "V1.16.1",
                    "NationalProfileOwner": "XX",
                    "NationalProfileVersion": "v1.0",
                },
            },
            "Payload": {
                "ResponsePayload": {
                    "ErrorInformation": {
                        "ErrorCode": 3007,
                        "ErrorDescription": error_message,
                    }
                }
            },
        }

    def _copy_header_from_request(self, request: dict) -> dict:
        header = deepcopy(request["Header"])
        header["Timestamp"] = etsi_datetime()
        return header

    def _generate_action_error(
        self, action_id: int, error_code: int, message: str
    ) -> dict:
        return {
            "ActionIdentifier": action_id,
            "ErrorInformation": {"ErrorCode": error_code, "ErrorDescription": message},
        }

    def _generate_create_response(
        self, action_id, object_id, updated_object=None
    ) -> dict:
        result = {
            "ActionIdentifier": action_id,
            "CREATEResponse": {"Identifier": object_id},
        }
        if updated_object is not None:
            result["CREATEResponse"]["HI1Object"] = updated_object
        return result

    def _parse_message(self, request: dict, context: dict = {}) -> dict:
        context["request"] = request
        errors = self.validator.validate(request)
        if len(errors) > 0:
            error_str = ""
            for error in errors:
                error_str += str(error)
            return self._generate_top_level_error(error_str)
        given_receiver_id = request["Header"]["ReceiverIdentifier"]
        if (not self.allowAnyReceiverID) and (
            (given_receiver_id["UniqueIdentifier"] != self.receiverID)
            or (given_receiver_id["CountryCode"] != self.receiverCC)
        ):
            return self._generate_top_level_error(
                f"Incorrect ReceiverID given: {given_receiver_id}"
            )
        response = {
            "Header": self._copy_header_from_request(request),
            "Payload": {"ResponsePayload": {"ActionResponses": {"ActionResponse": []}}},
        }
        action_responses = response["Payload"]["ResponsePayload"]["ActionResponses"][
            "ActionResponse"
        ]
        next_action_id = 0
        for action_request in request["Payload"]["RequestPayload"]["ActionRequests"][
            "ActionRequest"
        ]:
            action_id = action_request["ActionIdentifier"]
            if action_id != next_action_id:
                action_responses.append(
                    self._generate_action_error(
                        3007,
                        action_id,
                        f"Action request {action_id} is out-of-order, expecting {next_action_id}",
                    )
                )
                continue
            action_responses.append(self._parse_action(action_request, context))
        return response

    def _parse_action(self, action: dict, context: dict = {}) -> dict:
        action_identifier = action["ActionIdentifier"]
        action_type = list(action.keys())[1]
        if action_type != "CREATE":
            return self._generate_action_error(
                action_identifier, 3001, f"Action type {action_type} not supported"
            )
        obj = action["CREATE"]["HI1Object"]
        if (
            obj["@xsi:type"]
            != "{http://uri.etsi.org/03120/common/2020/09/Task}LDTaskObject"
        ):
            return self._generate_action_error(
                action_identifier, 3001, f"Object type {obj['@xsi:type']} not supported"
            )
        request_details = obj["task:RequestDetails"]
        delivery_url = None
        if obj.get("task:DeliveryDetails") is not None:
            dests = obj["task:DeliveryDetails"].get('task:DeliveryDestination')
            if dests is not None:
                if len(dests) != 1:
                    return self._generate_action_error(action_identifier, 3001, f'Expecting 1 task:DeliveryDestination, got {len(dests)}')
            addr = dests[0].get('task:DeliveryAddress')
            if addr is None:
                return self._generate_action_error(action_identifier, 3001, f'DeliveryDestination is missing a task:DeliveryAddress')
            url = addr.get('task:URL')
            if url is None:
                return self._generate_action_error(action_identifier, 3001, f'task:DeliveryAddress is not a task:URL')
            delivery_url = url
        if obj.get('task:Reference') is None:
                return self._generate_action_error(action_identifier, 3007, f'task:Reference not specified')
        task_reference = obj["task:Reference"]
        request_type = request_details["task:Type"]
        if (
            request_type["common:Owner"] != "ETSI"
            or request_type["common:Name"] != "TS103976RequestType"
        ):
            return self._generate_action_error(
                action_identifier,
                3001,
                f"Request type from unknown dictionary {request_type} not supported.",
            )
        query_type = request_type["common:Value"]
        try:
            if query_type == "VINtoCommsID":
                result = self._fulfil_VINtoCommsID(action, context)
                if result.errorCode != 0:
                    return self._generate_action_error(
                        action_identifier, result.errorCode, result.errorDescription
                    )
                else:
                    self._deliver_response(result, context['request'], task_reference, delivery_url)
                    return self._generate_create_response(
                        action_identifier, obj["ObjectIdentifier"], result.updatedObject
                    )
            elif query_type == "CommsIDtoVIN":
                result = self._fulfil_CommsIDToVIN(action, context)
                if result.errorCode != 0:
                    return self._generate_action_error(
                        action_identifier, result.errorCode, result.errorDescription
                    )
                else:
                    self._deliver_response(result, context['request'], task_reference, delivery_url)
                    return self._generate_create_response(
                        action_identifier, obj["ObjectIdentifier"], result.updatedObject
                    )

            # elif query_type == "VINtoLocationRecord":
            #     return {}}
            else:
                return self._generate_action_error(
                    action_identifier,
                    3007,
                    f"Invalid request type {query_type} - see TS 103 976 table 6.2.5-1 for valid values",
                )
        except Exception as ex:
                return self._generate_action_error(
                    action_identifier,
                    3007,
                    f"Failed delivering results ({ex})",
                )


    def _complete_LDRequest(self, ld_request_object: dict) -> dict:
        ld_request_object["LDTaskStatus"] = {
            "common:Owner": "ETSI",
            "common:Name": "LDTaskStatus",
            "common:Value": "Disclosed",
        }
        return ld_request_object

    def _fulfil_VINtoCommsID(
        self, action: dict, context: dict = {}
    ) -> FulfilmentOutcome:
        # TBD - fail if you specify a start or end time
        obj = action["CREATE"]["HI1Object"]
        request_details = obj["task:RequestDetails"]
        request_values = request_details["task:RequestValues"]["task:RequestValue"]
        if len(request_values) != 1:
            return FulfilmentOutcome(
                3007,
                f"Unexpected number of RequestValues ({len(request_values)}) for VINtoCommsID query, expecting 1",
            )
        request_value = request_values[0]
        request_value_format = request_value["task:FormatType"]
        if (
            request_value_format["task:FormatOwner"] != "ETSI"
            or request_value_format["task:FormatName"] != "VIN"
        ):
            return FulfilmentOutcome(
                3007,
                f"Unexpected FormatType ({request_value_format}) for VINtoCommsID query, expecting Owner 'ETSI' and FormatType 'VIN'",
            )
        vin = str(request_value["task:Value"])
        if not self.VIN_regex.match(vin):
            return FulfilmentOutcome(
                3007,
                f"VIN ({vin}) does not match TS 103 280 clause 6.57 regex ({self.VIN_regex.pattern})",
            )
        if vin[-1].isdigit():
            expected_responses = int(vin[-1])
        else:
            expected_responses = 0
        seed = int(sha1(vin.encode("ASCII")).hexdigest(), 16)
        random.seed(seed)
        records = []
        for i in range(expected_responses):
            records.append(
                {
                    "CommsID": {
                        "IMSI": f"99999{"".join([str(random.randint(0,9)) for i in range(10)])}"
                    }
                }
            )
        updatedObject = self._complete_LDRequest(obj)
        return_dict = {
            "VINtoCommsIDRecords" : records
        }
        return FulfilmentOutcome(updatedObject=updatedObject, records=return_dict)

    def _fulfil_CommsIDToVIN(self, action: dict, context: dict = {}
    ) -> FulfilmentOutcome:
        # TBD - fail if you specify a start or end time
        obj = action["CREATE"]["HI1Object"]
        request_details = obj["task:RequestDetails"]
        request_values = request_details["task:RequestValues"]["task:RequestValue"]
        if len(request_values) != 1:
            return FulfilmentOutcome(
                3007,
                f"Unexpected number of RequestValues ({len(request_values)}) for VINtoCommsID query, expecting 1",
            )
        request_value = request_values[0]
        request_value_format = request_value["task:FormatType"]
        if (
            request_value_format["task:FormatOwner"] != "ETSI"
            or not request_value_format["task:FormatName"] in RPS.SUPPORTED_COMMS_IDS
        ):
            return FulfilmentOutcome(
                3007,
                f"Unexpected FormatType ({request_value_format}) for VINtoCommsID query, expecting Owner 'ETSI' and FormatType from {RPS.SUPPORTED_COMMS_IDS}",
            )
        commsID = str(request_value["task:Value"])
        expected_responses = 1
        seed = int(sha1(commsID.encode("ASCII")).hexdigest(), 16)
        random.seed(seed)
        records = []
        for i in range(expected_responses):
            records.append(
                {
                    "VIN": random.choice(RPS.ALLOWED_VINS)
                }
            )
        updatedObject = self._complete_LDRequest(obj)
        return_dict = {
            "CommsIDtoVINRecords" : records
        }
        return FulfilmentOutcome(updatedObject=updatedObject, records=return_dict)
        

    def _deliver_response(self, fulfilment_outcome, original_request, task_reference, url):
        deliver_id = str(uuid4())
        deliver_request = {
            "Header": self._copy_header_from_request(original_request),        
            "Payload": {"RequestPayload": {"ActionRequests": {"ActionRequest": [ { 
                "ActionIdentifier": 0,
                "DELIVER" :
                {
                    "Identifier" : deliver_id,
                    "HI1Object" : {
                        "@xsi:type" : "{http://uri.etsi.org/03120/common/2019/10/Delivery}DeliveryObject",
                        "ObjectIdentifier" : deliver_id,
                        "delivery:Reference" : {
                            "delivery:LDID" : task_reference,
                        },
                        "delivery:Manifest" : {
                            "delivery:Specification" : {
                                "common:Owner" : "ETSI",
                                "common:Name" : "ManifestSpecification",
                                "common:Value" : "TS103976"
                            }
                        },
                        "delivery:Delivery" : {
                            "delivery:JSONData" : fulfilment_outcome.records
                        }
                    }
                }
            }
            ]}}}}
        
        deliver_request['Header']['SenderIdentifier'], deliver_request['Header']['ReceiverIdentifier'] = deliver_request['Header']['ReceiverIdentifier'], deliver_request['Header']['SenderIdentifier']

        deliver_request['Header']['@xmlns:delivery'] = "http://uri.etsi.org/03120/common/2019/10/Delivery"
        deliver_request['Header']['TransactionIdentifier'] = str(uuid4())
        if ('@xmlns:task') in deliver_request: 
            deliver_request['Header'].pop('@xmlns:task')

        if self.alwaysLogResults:
            print(dumps(deliver_request))
            
        if url is None:
            logging.warn("No delivery destination specified")
            match self.behaviourIfNoDeliveryDest:
                case DeliveryBehaviourEnum.SKIPDELIVERY:
                    logging.info("Skipping delivery")
                    return
                case DeliveryBehaviourEnum.FAIL:
                    logging.warn("Raising an exception due to DeliveryBehaviour set to FAIL")
                    raise Exception ("No delivery destination specified")
                case DeliveryBehaviourEnum.DELIVERTODEFAULT:
                    if self.defaultDeliveryDest is not None:
                        logging.info(f"Defaulting to default delivery destination {self.defaultDeliveryDest}")
                        url = self.defaultDeliveryDest
                        # fall out of match and continue delivery               
                    else:
                        logging.error(f"behaviourIfNoDeliveryDest is DELIVERTODEFAULT but no default delivery destination was set. Skipping delivery.")
                        return
                case DeliveryBehaviourEnum.WRITETOSTDOUT:
                    logging.info(f"Delivering to STDOUT")
                    print(dumps(deliver_request))
                    return
                case DeliveryBehaviourEnum.WRITETOSTDERR:
                    logging.info(f"Delivering to STDERR")
                    print(dumps(deliver_request), file=stderr)
                    return
                case _:
                    logging.error(f"Unexpected behaviourIfNoDeliveryDest set ({self.behaviourIfNoDeliveryDest}). Skipping delivery")
                    return
                

        errors = self.validator.validate(deliver_request)
        if len(errors) > 0:
            logging.error(f"Vailed validating delivery: {errors}")
            raise Exception("Error validating delivery")
        r = post(url, dumps(deliver_request), timeout=self.deliveryTimeout) # TBD - configure timeout and error handling
        r.close()


    def respond_to(self, request: dict, context: any = {}) -> dict:
        response = self._parse_message(request, context)
        own_errors = self.validator.validate(response)
        if len(own_errors) > 0:
            logging.error(f"Failed validating own response: {own_errors}")
        return response



import logging
import argparse
import sys
import json

from .lea import LEA
from .rps import RPS, DeliveryBehaviourEnum

from .validator import TS103120Validator, TS103976Validator

def lea_console():
    parser = argparse.ArgumentParser(
        description="Generates JSON messages for simple example requests following TS 103 976, \
                                          encapsulated in TS 103 120. This is an example implementation for testing \
                                          and demonstration purposes, and should not be used in production environments.",

    )
    parser.add_argument("-s", "--sendercc", default="XX", help="Sender Country Code used in the TS 103 120 header")
    parser.add_argument("-u", "--senderid", default="SENDER", help="Sender Unique ID used in the TS 103 120 header")
    parser.add_argument("-r", "--receivercc", default="XX", help="Receiver Country Code used in the TS 103 120 header")
    parser.add_argument("-i", "--receiverid", default="RECEIVER", help="Receiver Unique ID used in the TS 103 120 header")
    parser.add_argument("-t", "--taskref", default="XX-1-234", help="Task Reference (LDID) used in the TS 103 120 LDTaskRequest")
    parser.add_argument("-d", "--deliveryurl", help="Delivery URL included in the TS 103 120 LDTaskRequest, as a destination for results")

    group = parser.add_mutually_exclusive_group()
    group.add_argument("-v", "--vin", help="VIN to query as a RequestValue in the TS 103 120 LDTaskRequest")
    group.add_argument("-m", "--imsi", help="IMSI to query as a RequestValue in the TS 103 120 LDTaskRequest")

    args = parser.parse_args()

    lea = LEA(args.sendercc, args.senderid, args.deliveryurl)

    if (args.vin is not None):        
        json_doc = lea.generate_vin_to_comms_id(
            args.receivercc, args.receiverid, args.taskref, args.vin
        )
    elif (args.imsi is not None):
        json_doc = lea.generate_comms_id_to_vin(
            args.receivercc, args.receiverid, args.taskref, args.imsi
        )
    else:
        json_doc = lea.generate_vin_to_comms_id(
            args.receivercc, args.receiverid, args.taskref, "1G9Y817H34LSP7293"
        )

    print(json.dumps(json_doc))


def rps_console():
    parser = argparse.ArgumentParser(
        description="Responds to JSON messages for simple example requests following TS 103 976, \
                     encapsulated in TS 103 120, and produces simple example responses. \
                     TS 103 120 response messages will be written to stdout. The DELIVER message \
                     containing any results will either be set to the delivery address specified in the LDTaskRequest \
                     or written to disk\
                     This is an example implementation for testing \
                     and demonstration purposes, and should not be used in production environments.",
)

    parser.add_argument("-r", "--receivercc", default="XX", help="Receiver Country Code for the TS 103 120 receiver. Messages with other Receiver Country Code values will be rejected unless the '--allowanyid' flag is set.")
    parser.add_argument("-i", "--receiverid", default="RECEIVER", help="Receiver Unique ID for the TS 103 120 receiver. Messages with other Receiver Unique ID values will be rejected unless the '--allowanyid' flag is set.")
    parser.add_argument("-a", "--allowanyid", default=False, help="Respond to messages to any Receiver ID, regardless of the values set by '--receivercc' or '--receiverid'.")
    parser.add_argument("-d", "--delivery", type=lambda b: DeliveryBehaviourEnum.from_string(b), choices=list(DeliveryBehaviourEnum), default=DeliveryBehaviourEnum.SKIPDELIVERY, help="Behaviour if a request does not specify a delivery address")
    parser.add_argument("--defaultaddress", default=None, help="Default delivery address. Implies --delivery delivertodefault")
    parser.add_argument("-j", "--input", type=argparse.FileType("r"), default=sys.stdin, help="Path to input file (if absent, stdin is used)")
    parser.add_argument("-v", "--verbose", action="count", help="Verbosity. Can be specified multiple times for extra verbosity")

    args = parser.parse_args()

    match args.verbose:
        case None:
            logging.basicConfig(level=logging.ERROR)
        case 1:
            logging.basicConfig(level=logging.WARN)
        case 2:
            logging.basicConfig(level=logging.INFO)
        case 3:
            logging.basicConfig(level=logging.DEBUG)

    logging.debug(args)
    
    instance_doc = json.loads(args.input.read())
    args.input.close()
    logging.debug("Input:")
    logging.debug(instance_doc)

    logging.debug("Creating RPS instance")
    rps = RPS(args.receivercc, args.receiverid, args.allowanyid, args.delivery, args.defaultaddress)
    
    logging.debug("Responding to input")
    response = rps.respond_to(instance_doc)
    
    logging.debug("Response from RPS")
    print(json.dumps(response))


def search_key(data, key, path=""):
    if type(data) is  dict:
        for k, v in data.items():
            path="{0} -> {1}".format(path, k)
            if k == key or v == key:
                return (k, v, path)
            res = search_key(data[k], key, path)
            if res is not None:
                return res
    if type(data) is list:
        for v in data:
            res = search_key(v, key, path)
            if res is not None:
                return res

def lea_parse():
    parser = argparse.ArgumentParser()
    parser.add_argument("-j", "--input", type=argparse.FileType("r"), default=sys.stdin, help="Path to input file (if absent, stdin is used)")

    args = parser.parse_args()
    instance_doc = args.input.read()
    json_doc = json.loads(instance_doc)
    args.input.close()

    validator = TS103120Validator()
    errors = validator.validate(json_doc)
    if len(errors) > 0:
        for error in errors:
            print(error)
    else:
        print ("TS103120 validation OK")

    # quick and dirty
    _, json_data, _ = search_key(json_doc, 'delivery:JSONData')
    if json_data is None:
        raise Exception('No delivery:JSONData key found')

    inner_validator = TS103976Validator()
    errors = inner_validator.validate(json_data)
    if len(errors) > 0:
        for error in errors:
            print(error)
    else:
        print ("TS103976 validation OK")

    print ("Done")
   